/* $Id: sfskeyfetch.C,v 1.13 1999/05/07 09:06:01 dm Exp $ */

/*
 *
 * Copyright (C) 1999 Michael Kaminsky (kaminsky@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfskey.h"

void
keyfetch::finish (int err)
{
  (*cb) (err);
  delete this;
}

void
keyfetch::bfinish (bool b)
{
  finish (!b);
}

void
keyfetch::dosymlink (str name, str contents, bool b)
{
  if (!b) 
    bfinish (b);
  else {
    makesymlink (agentc, name, contents, wrap (this, &keyfetch::bfinish));
  }
}

void
keyfetch::dofetch_cb (sfsauth_fetchres *resp, clnt_stat err)
{
  if (err) {
    fatal << "authserv: " << err << "\n";
  }

  switch (resp->status) {
  case SFSAUTH_SRP_OK:
    break;
  case SFSAUTH_FETCH_FAILED:
    warn << "Unable to fetch private information.\n";
    delete resp;
    finish (1);
    return;
  default:
    warn << "Unknown response from authserv during fetch.\n";
    delete resp;
    finish (1);
    return;
  }

#if 0
  warn << "seckey = " << resp->resok->seckey 
       << "; hostid = " << resp->resok->hostid[0] << "\n";
#endif

  ptr<rabin_priv> sk = import_rabin_decrypt_sec (resp->resok->seckey, &eksb);
  if (!sk) {
    warn << "Could not decrypt private key!\n";
    delete resp;
    finish (1);
    return;
  }

  str path;
  str user = username << "@" << authserv;

  if (ignorehostid) {
    warn << "Remote host doesn't match the one we asked for!  "
	 << "Ignoring host id\n";
  }
  else {
    path = authserv << ":" << armor32 (str (resp->resok->hostid.base (), 
					    sizeof (sfs_hash)));
    warnx << "Authserv is " << path << "\n";
  }

  delete resp;

  if (keyfile) {
    if (str2file (keyfile, export_rabin_priv (*sk, pwd, user), 0600))
      finish (0);
    else 
      finish (errno);
  }
  else {
    if (ignorehostid)
      addkey (agentc, export_rabin_priv (*sk, NULL, user),
	      wrap (this, &keyfetch::bfinish), true);
    else
      addkey (agentc, export_rabin_priv (*sk, NULL, user),
	      wrap (this, &keyfetch::dosymlink, authserv, path), true);
  }
}

void
keyfetch::srp_phase2done (sfsauth_srp2res *resp, clnt_stat err)
{
  if (err) {
    fatal << "authserv: " << err << "\n";
  }

  switch (resp->status) {
  case SFSAUTH_SRP_OK:
    break;
  case SFSAUTH_SRP_PHASE2:
    warn << "Authserv aborted SRP in Phase 2--probably wrong password.\n";
    delete resp;
    finish (1);
    return;
  default:
    warn << "Unknown response from authserv in Phase 2.\n";
    delete resp;
    finish (1);
    return;
  }

  /* Check proof from server */
  if (!srpclnt.phase3clnt (reinterpret_cast<u_char *> 
			   (resp->resok->servproof.base ()))) {
    fatal << "Server's proof of secret K does not check!\n";
  }

  warn << "SRP secret key negotiation sucessfull!\n";
#if 0
  warn << "Session Key = " << srpclnt.S << "\n";
#endif

  /* Encrypt the underlying transport */
  sfsauth_client_crypt (authclnt, srpclnt.S);

  /* Now we can safely retrieve the private information */
  sfsauth_fetchres *resp2 = New sfsauth_fetchres;
  authclnt->call (SFSAUTHPROC_FETCH, NULL, resp2,
		  wrap (this, &keyfetch::dofetch_cb, resp2));

  delete resp;
}

void
keyfetch::srp_phase1done (sfsauth_srp1res *resp, clnt_stat err)
{
  if (err) {
    fatal << "authserv: " << err << "\n";
  }

  switch (resp->status) {
  case SFSAUTH_SRP_OK:
    break;
  case SFSAUTH_SRP_PHASE1:
    warn << "Authserv aborted SRP in Phase 1.\n";
    delete resp;
    finish (1);
    return;
  default:
    warn << "Unknown response from authserv in Phase 1.\n";
    delete resp;
    finish (1);
    return;
  }

#if 0
  warn << "salt = " << resp->resok->salt << "; B = " << resp->resok->B
       << "; u = " << resp->resok->u << "\n";
#endif

  str host = pw_getptext (resp->resok->salt);
  if (authserv != host)
    ignorehostid = true;
  else
    ignorehostid = false;

  bigint x = pw_getint (pwd, resp->resok->salt, 
			srpclnt.N.nbits () - 1, &eksb);
  if (!x) {
    warn << "SRP modulus too short\n";
    delete resp;
    finish (1);
    return;
  }

  if (!srpclnt.phase2clnt (x, bigint (resp->resok->B, 16), 
			   bigint (resp->resok->u, 16), resp->resok->salt, 
			   username)) {
    warn << "Client aborted SRP in Phase 2.\n";
    delete resp;
    finish (1);
    return;
  }

  /* Send proof of key */
  sfsauth_srp2arg arg;
  memcpy (arg.clntproof.base (), srpclnt.proof, arg.clntproof.size ());

  sfsauth_srp2res *resp2 = New sfsauth_srp2res;
  authclnt->call (SFSAUTHPROC_SRP_PHASE2, &arg, resp2,
		  wrap (this, &keyfetch::srp_phase2done, resp2));

  delete resp;
}

void
keyfetch::srp_phase0done (sfsauth_srp0res *resp, clnt_stat err)
{
  if (err) {
    fatal << "authserv: " << err << "\n";
  }

  switch (resp->status) {
  case SFSAUTH_SRP_OK:
    break;
  case SFSAUTH_SRP_NOTTHERE:
    warn << "Authserv doesn't know about user " << username << ".\n";
    delete resp;
    finish (1);
    return;
  case SFSAUTH_SRP_BADSRPINFO:
    warn << "Authserv doesn't have SRP information for " << username << ".\n";
    delete resp;
    finish (1);
    return;
  case SFSAUTH_SRP_PHASE0:
    warn << "Authserv aborted SRP in Phase 0.\n";
    delete resp;
    finish (1);
    return;
  default:
    warn << "Unknown response from authserv in Phase 0.\n";
    delete resp;
    finish (1);
    return;
  }

  if (!srpclnt.set_params (bigint (resp->resok->prime, 16), 
			   bigint (resp->resok->generator, 16))) {
    warn << "SRP parameters N and g failed to check out!\n";
    delete resp;
    finish (1);
    return;
  }

  if (!srpclnt.phase1clnt ()) {
    warn << "Failied to negotiate a session key via SRP (Phase 1).\n";
    delete resp;
    finish (1);
    return;
  }

  sfsauth_srp1arg arg;
  arg.A = srpclnt.A.getstr (16);

  sfsauth_srp1res *resp2 = New sfsauth_srp1res;
  authclnt->call (SFSAUTHPROC_SRP_PHASE1, &arg, resp2,
		  wrap (this, &keyfetch::srp_phase1done, resp2));

  delete resp;
}

void
keyfetch::gotconres (int fd, sfs_connectres *conresp, clnt_stat err)
{
  if (err) {
    warn << authserv << ": " << err << "\n";
    finish (EIO);
    return;
  }
  if (conresp->status) {
    warn << authserv << ": " << conresp->status << "\n";
    switch (conresp->status) {
    default:
    case SFS_NOSUCHHOST:
      finish (ENOENT);
      return;
    case SFS_NOTSUPP:
      finish (EPROTONOSUPPORT);
      return;
    case SFS_TEMPERR:
      finish (EAGAIN);
      return;
    }
  }

  authclnt = aclnt::alloc (authxprt, sfsauth_program_1);

  warnx << "Executing SRP protocol.  This may take a few seconds.\n";

  sfsauth_srp0arg arg;
  arg.username = username;

  sfsauth_srp0res *resp = New sfsauth_srp0res;
  authclnt->call (SFSAUTHPROC_SRP_PHASE0, &arg, resp,
		  wrap (this, &keyfetch::srp_phase0done, resp));
}

void
keyfetch::connected (int fd)
{
  if (fd < 0) {
    warn << "Could not contact host " << authserv << "\n";
    finish (errno);
    return;
  }

  tcp_nodelay (fd);
  authxprt = axprt_crypt::alloc (fd);
  ptr<aclnt> c = aclnt::alloc (authxprt, sfs_program_1);

  sfs_connectarg arg;
  arg.release = SFS_RELEASE;
  arg.service = SFS_AUTHSERV;
  arg.name = authserv;

  sfs_connectres *conresp = New sfs_connectres;

  c->call (SFSPROC_CONNECT, &arg, conresp,
	   wrap (this, &keyfetch::gotconres, fd, conresp));
}

void
keyfetch::gotpwd (str password)
{
  if (password == "") {
    warnx ("Aborted.\n");
    finish (1);
    return;
  }

  pwd = password;

  tcpconnect (authserv, sfs_port, wrap (this, &keyfetch::connected));
}

keyfetch::keyfetch (str kf, str as, str un, cb_t cb)
  : keyfile (kf), authserv (as), username (un), cb (cb)
{
  if (!getkbdpwd ("Passphrase for " << username << "@" << authserv
		  << ": ", &rnd_input, 
		  wrap (this, &keyfetch::gotpwd)))
    fatal ("stdin must be a tty.\n");
}
