/* $Id: sfskeyreg.C,v 1.17 1999/05/13 20:28:37 kaminsky Exp $ */

/*
 *
 * Copyright (C) 1999 Michael Kaminsky (kaminsky@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfskey.h"

void
keyregister::finish (int err)
{
  (*cb) (err);
  delete this;
}

void
keyregister::doregister_cb (sfsauth_registerres *_resp, clnt_stat err)
{
  auto_ptr<sfsauth_registerres> resp (_resp);

  if (err) {
    fatal << "authserv: " << err << "\n";
  }

  switch (*resp) {
  case SFSAUTH_REGISTER_OK:
    warn << "authserv: Registration successfull...installing key:\n";
    warnx << username << ":0x" << key->modulus ().getstr (16) << "\n";
    finish (0);
    return;
  case SFSAUTH_REGISTER_NOTSOCK:
    warn << "authserv: Registration failed--can't conect via network.\n";
    finish (1);
    return;
  case SFSAUTH_REGISTER_BADUSERNAME:
    warn << "authserv: Registration failed--username not in password file.\n";
    finish (1);
    return;
  case SFSAUTH_REGISTER_WRONGUID:
    warn << "authserv: Registration failed--uid doesn't match username.\n";
    finish (1);
    return;
  case SFSAUTH_REGISTER_DENYROOT:
    warn << "authserv: Registration failed--Can't register root account.\n";
    finish (1);
    return;
  case SFSAUTH_REGISTER_BADSHELL:
    warn << "authserv: Registration failed--shell is not in /etc/shells.\n";
    finish (1);
    return;
  case SFSAUTH_REGISTER_DENYFILE:
    warn << "authserv: Registration failed--user explicitly denied.\n";
    finish (1);
    return;
  case SFSAUTH_REGISTER_BADPASSWORD:
    warn << "authserv: Registration failed--incorrect password.\n";
    finish (1);
    return;
  case SFSAUTH_REGISTER_USEREXISTS:
    warn << "authserv: Registration failed--user exists in auth database.\n";
    finish (1);
    return;
  case SFSAUTH_REGISTER_FAILED:
    warn << "authserv: Registration failed--could not register new user.\n";
    finish (1);
    return;
  default:
    warn << "Unknown response from authserv\n";
    finish (1);
    return;
  }
}

void
keyregister::gotsrpparams (sfsauth_srpinfo *resp, clnt_stat err)
{
  if (err) {
    fatal << "authserv: " << err << "\n";
  }

  sfsauth_registerarg arg;
  arg.username = username;
  arg.password = unixpwd;
  arg.pubkey = key->modulus ();

  /* Get private key parameters, if needed */
  if (opt_sec) {
    srp srpclnt;

    if (!srpfile) {
      if (!srpclnt.set_params (bigint (resp->prime, 16),
			       bigint (resp->generator, 16)))
	fatal << "SRP parameters N and g failed to check out!\n";
    }
    else {
      srpclnt.set_params (bigint (resp->prime, 16),
			  bigint (resp->generator, 16), false);
    }
    
    str authhostname = myname ();
    if (!authhostname)
      fatal << "Could not determine our own (i.e. the authserv's) hostname\n";
    str salt = pw_gensalt (SK_SRP_EKSBF_COST, authhostname);

    eksblowfish eksb;
    bigint x = pw_getint (pwd, salt, srpclnt.N.nbits () - 1, &eksb);
    if (!x)
      fatal << "SRP modulus too short\n";

    arg.srpinfo.alloc ();
    arg.srpinfo->prime = resp->prime;
    arg.srpinfo->generator = resp->generator;
    arg.srpinfo->salt = salt;
    arg.srpinfo->verifier = srpclnt.calc_verifier (x).getstr (16);
    arg.srpinfo->seckey = export_rabin_encrypt_sec (*key, &eksb);
    if (!arg.srpinfo->seckey)
      fatal << "Could not encrypt secret key\n";
  }

  delete resp;

  sfsauth_registerres *resp2 = New sfsauth_registerres;
  authclnt->call (SFSAUTHPROC_REGISTER, &arg, resp2,
		  wrap (this, &keyregister::doregister_cb, resp2));
}

void
keyregister::gotpwdunix (str _unixpwd)
{
  /* Empty string => quit */
  if (_unixpwd == "") {
    warnx ("Aborted.\n");
    finish (1);
    return;
  }

  username = getlogin ();
  unixpwd = _unixpwd;

  /* Do the RPC to the authserv */
  fd = suidgetfd ("authserv");
  
  if (fd < 0)
    fatal << "Could not find 'authserv' unix domain socket.\n";

  authclnt = aclnt::alloc (authxprt = axprt_stream::alloc (fd),
                           sfsauth_program_1);

  sfsauth_srpinfo *resp = New sfsauth_srpinfo;

  if (opt_sec) {
    warnx << "Checking SRP parameters.  This may take a few seconds.\n";

    if (!pwd)
      fatal << "Private key has no password--can't install on authserv\n";

    if (!srpfile)
      authclnt->call (SFSAUTHPROC_SRP_GETGLOBALPARAMS, NULL, resp,
		      wrap (this, &keyregister::gotsrpparams, resp));
    else {
      ptr<srp> s = import_srp_params (file2str (srpfile));
      if (!s)
	fatal << "Could not load SRP parameters from " << srpfile << "\n";
      resp->prime = s->N.getstr (16);
      resp->generator = s->g.getstr (16);
      gotsrpparams (resp, RPC_SUCCESS);
    }
  }
  else {
    gotsrpparams (resp, RPC_SUCCESS);
  }
}

void
keyregister::gotpwdkey (str _pwd)
{
  pwd = _pwd;
  key = import_rabin_priv (key_asc, pwd);

  if (!key)
    fatal ("could not decode secret key\n");

  if (!getkbdpwd ("UNIX Password: ", &rnd_input, 
		  wrap (this, &keyregister::gotpwdunix)))
    fatal ("stdin must be a tty.\n");
}

keyregister::keyregister (str kf, str srpf, bool sec, cb_t cb)
  : keyfile (kf), srpfile (srpf), opt_sec (sec), cb (cb)
{
  key_asc = file2wstr (keyfile);
  if (!key_asc)
    fatal ("%s: %m\n", keyfile.cstr ());

  if (import_rabin_priv_need_pwd (key_asc)) {
    if (!getkbdpwd ("Passphrase for " << keyfile << ": ",
		    &rnd_input, wrap (this, &keyregister::gotpwdkey)))
      fatal ("stdin must be a tty.\n");
  }
  else {
    gotpwdkey (NULL);
  }
}
