// -*-c++-*-
/* $Id: bitvec.h,v 1.3 1999/02/18 06:58:43 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/* XXX - this is suboptimal */

#ifndef _ASYNC_BITVEC_H_
#define _ASYNC_BITVEC_H_ 1

#include "str.h"

class bitvec {
  friend void swap (bitvec &a, bitvec &b);
  enum { bitsperdat = 8 * sizeof (int) };

  int *dat;
  size_t nbits;

  void init () { dat = NULL; nbits = 0; }
  static size_t nbytes (size_t nb)
    { return ((nb + bitsperdat - 1) / bitsperdat) * sizeof (int); }
  void datalloc (size_t nb) {
    if (nb)
      dat = static_cast<int *> (xrealloc (dat, nbytes (nb)));
    else {
      /* Jump through hoops for dmalloc */
      xfree (dat);
      dat = NULL;
    }
  }

#ifdef CHECK_BOUNDS
#define bcheck(n) assert ((size_t) (n) < nbits)
#define rcheck(s, e) assert (s <= e && e <= nbits)
#else /* !CHECK_BOUNDS */
#define bcheck(n)
#define rcheck(s, e)
#endif /* !CHECK_BOUNDS */

  class wbit {
    friend class bitvec;

    int *const intp;
    const size_t bitpos;
    const bool bit;

    wbit (int *i, size_t b) : intp (i), bitpos (b), bit (*i & 1 << b) {}

  public:
    operator bool () const { return bit; }
    wbit &operator= (bool v) {
      if (v)
        *intp |= 1 << bitpos;
      else
        *intp &= ~(1 << bitpos);
      return *this;
    }
  };

  void range_set (size_t s, size_t e) {
    size_t sp = s / bitsperdat, ep = e / bitsperdat;
    int sb = s % bitsperdat, eb = e % bitsperdat;
    if (sp == ep) {
      if (eb)
	dat[sp] |= ((unsigned) -1 << sb) & ~((unsigned) -1 << eb);
    }
    else {
      dat[sp] |= ((unsigned) -1 << sb);
      if (eb)
	dat[ep] |= ~((unsigned) -1 << eb);
      memset (&dat[sp+1], 0xff, (ep - sp - 1) * sizeof (int));
    }
  }

  void range_clr (size_t s, size_t e) {
    size_t sp = s / bitsperdat, ep = e / bitsperdat;
    int sb = s % bitsperdat, eb = e % bitsperdat;
    if (sp == ep) {
      if (eb)
	dat[sp] &= ~((unsigned) -1 << sb) | ((unsigned) -1 << eb);
    }
    else {
      dat[sp] &= ~((unsigned) -1 << sb);
      if (eb)
	dat[ep] &= ((unsigned) -1 << eb);
      bzero (&dat[sp+1], (ep - sp - 1) * sizeof (int));
    }
  }

public:
  bitvec () { init (); }
  bitvec (size_t n) { init (); zsetsize (n); }
  bitvec (const bitvec &v) { init (); *this = v; }
  ~bitvec () { xfree (dat); }

  void setsize (size_t n) { datalloc (nbits = n); }
  void zsetsize (size_t n) {
    datalloc (n);
    if (n > nbits)
      range_clr (nbits, n);
    nbits = n;
  }
  size_t size () const { return nbits; }

  bool at (ptrdiff_t i) const {
    bcheck (i);
    return dat[(size_t) i / bitsperdat] & 1 << ((size_t) i % bitsperdat);
  }
  bool operator[] (ptrdiff_t i) const {
    bcheck (i);
    return dat[(size_t) i / bitsperdat] & 1 << ((size_t) i % bitsperdat);
  }
  wbit operator[] (ptrdiff_t i) {
    bcheck (i);
    return wbit (dat + (size_t) i / bitsperdat, (size_t) i % bitsperdat);
  }
  void setrange (size_t s, size_t e, bool v) {
    rcheck (s, e);
    if (v)
      range_set (s, e);
    else
      range_clr (s, e);
  }

  bitvec &operator= (const bitvec &v) {
    setsize (v.nbits);
    memcpy (dat, v.dat, nbytes (nbits));
    return *this;
  }

#undef bcheck
#undef rcheck
};

inline void
swap (bitvec &a, bitvec &b)
{
  int *dat = a.dat;
  a.dat = b.dat;
  b.dat = dat;

  size_t nbits = a.nbits;
  a.nbits = b.nbits;
  b.nbits = nbits;
}

inline const strbuf &
strbuf_cat (const strbuf &sb, const bitvec &v)
{
  char *p = sb.tosuio ()->getspace (v.size ());
  for (size_t i = 0; i < v.size (); i++)
    p[i] = v[i] ? '1' : '0';
  sb.tosuio ()->print (p, v.size ());
  return sb;
}

#endif /* !_ASYNC_BITVEC_H_ */
