/* $Id: sfscd.C,v 1.27 1999/05/17 07:02:20 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */


#include "sfscd.h"
#include <pwd.h>
#include <grp.h>

#include "parseopt.h"

bool nomounting;

ihash<const int32_t, cdaemon, &cdaemon::prog, &cdaemon::hlink> cdtab;

static const char *
stripdir (const char *s)
{
  const char *p = strrchr (s, '/');
  return p ? p + 1 : s;
}

cdaemon::cdaemon (int32_t prog)
  : prog (prog)
{
  cdtab.insert (this);
}

cdaemon::~cdaemon ()
{
  cdtab.remove (this);
}

void
cdaemon::eof ()
{
  warn << argv[0] << ": EOF\n";
  x = NULL;
  c = NULL;
  s = NULL;
  for (srvinfo *si = servers.first; si; si = servers.next (si))
    si->destroy (true);
}

bool
cdaemon::launch (bool sync)
{
  if (sync)
    x = axprt_unix_spawnv (argv[0], argv);
  else
    x = axprt_unix_aspawnv (argv[0], argv);
  if (x) {
    c = aclnt::alloc (x, sfscd_program_1);
    s = asrv::alloc (x, sfscdcb_program_1, wrap (this, &cdaemon::dispatch));

    sfscd_initarg ia;
    ia.name = name;
    c->call (SFSCDPROC_INIT, &ia, NULL, aclnt_cb_null);
  }
  else
    c = NULL;
  return c;
}

void
cdaemon::dispatch (svccb *sbp)
{
  if (!sbp) {
    eof ();
    return;
  }
  switch (sbp->proc ()) {
  case SFSCDCBPROC_NULL:
    sbp->reply (NULL);
    break;
  case SFSCDCBPROC_AGENTREQ:
    {
      sfscd_agentreq_arg *aa = sbp->template getarg<sfscd_agentreq_arg> ();
      if (aa->agentreq.type == AGENTCB_AUTHINIT) {
	  if (aa->agentreq.init->requestor.len ())
	    aa->agentreq.init->requestor
	      = name << "!" << aa->agentreq.init->requestor;
	  else
	    aa->agentreq.init->requestor = name;
      }
      if (usrinfo *ui = usrtab[aa->aid])
	ui->authreq (&aa->agentreq, sbp);
      else
	sbp->replyref (sfsagent_auth_res (false));
      break;
    }
  case SFSCDCBPROC_IDLE:
    sbp->reply (NULL);
    break;
  case SFSCDCBPROC_DELFS:
    srvinfo::destroy (*sbp->template getarg<nfspath3> (), this, false);
    sbp->reply (NULL);
    break;
  case SFSCDCBPROC_HIDEFS:
    //flushpath (*sbp->template getarg<nfspath3> ());
    srvinfo::show (*sbp->template getarg<nfspath3> (), this, false);
    sbp->reply (NULL);
    break;
  case SFSCDCBPROC_SHOWFS:
    srvinfo::show (*sbp->template getarg<nfspath3> (), this, true);
    sbp->reply (NULL);
    break;
  }
}

cdaemon *
cdaemon::lookup (int32_t prog, int32_t vers)
{
  cdaemon *cd;
  for (cd = cdtab[prog]; cd; cd = cdtab.nextkeq (cd))
    for (size_t i = 0; i < cd->vers.size (); i++)
      if (cd->vers[i] == vers)
	return cd;
  return NULL;
}

static void
parseconfig (str cf)
{
  parseargs pa (cf);
  bool errors = false;

  int line;
  vector<str> av;
  struct cdaemon *cd = NULL;
  qhash<str, int> dnotab;

  while (pa.getline (&av, &line)) {
    if (!strcasecmp (av[0], "Program")) {
      if (cd) {
	errors = true;
	warn << cf << ":" << line << ": expected Exec for Program "
	     << cd->prog << "\n";
	delete cd;
	cd = NULL;
      }
      int32_t prog;
      if (av.size () < 3 || !convertint (av[1], &prog)) {
	cd = NULL;
	errors = true;
	warn << cf << ":" << line << ": usage: Program <prog> <vers>\n";
	continue;
      }
      cd = New cdaemon (prog);
      for (size_t i = 2; i < av.size (); i++) {
	int32_t v;
	if (!convertint (av[i], &v)) {
	  errors = true;
	  warn << cf << ":" << line << ": version '"
	       << av[i] << "' should be a number.\n";
	}
	else if (cdaemon::lookup (cd->prog, v)) {
	  errors = true;
	  warn << cf << ":" << line << ": Redefinition of Program "
	       << cd->prog << " Version " << v << "\n";
	}
	else
	  cd->vers.push_back (v);
      }
    }
    else if (!strcasecmp (av[0], "Exec")) {
      if (av.size () < 2) {
	errors = true;
	warn << cf << ":" << line << ": path must follow Exec\n";
      }
      else if (cd) {
	if (av[1][0] == '/')
	  cd->argv.push_back (av[1]);
	else
	  cd->argv.push_back (strbuf () << execdir << "/" << av[1]);
	for (size_t i = 2; i < av.size (); i++)
	  cd->argv.push_back (av[i]);

	cd->name = stripdir (cd->argv[0]);
	if (!dnotab[cd->name])
	  dnotab.insert (cd->name, 0);
	cd->name = cd->name << "_" << ++*dnotab[cd->name];

	if (!cd->launch (true)) {
	  errors = true;
	  warn << cf << ":" << line << ": " << cd->argv[0]
	       << ": " << strerror (errno) << "\n";
	}
      }
      else {
	errors = true;
	warn << cf << ":" << line << ": No program for Exec\n";
      }
      cd = NULL;
    }
    else {
      errors = true;
      warn << cf << ":" << line << ": Unknown directive '"
	   << av[0] << "'\n";
    }
  }

  if (errors)
    fatal << "errors in configuration file\n";
}

static void
nullfn ()
{
}

static void
drop_privs (void)
{
  setuid (sfs_uid);
  if (chdir (sfsdir) < 0)
    fatal ("%s: %m\n", sfsdir.cstr ());
}

static void
init_ids ()
{
  setgid (sfs_gid);
  setgroups (0, NULL);
}

static void
usage ()
{
  warnx << "usage: " << progname << " [-f configfile] [-x]\n";
  exit (1);
}

extern char *optarg;
extern int optind;

int
main (int argc, char **argv)
{
  str configfile;
  setprogname (argv[0]);

  int ch;
  while ((ch = getopt (argc, argv, "xf:")) != -1)
    switch (ch) {
    case 'f':
      configfile = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if (argc > 1)
    usage ();

#ifdef MAINTAINER
  if (getenv ("SFS_NOMOUNTING"))
    nomounting = true;
#endif /* MAINTAINER */
  umask (0);
  sfsconst_init ();

  if (!configfile)
    configfile = sfsconst_etcfile_required ("sfscd_config");

  struct stat sb;
  if (stat (sfsroot, &sb) < 0 && errno == ENOENT && mkdir (sfsroot, 0555) < 0)
    fatal ("mkdir (%s): %m\n", sfsroot);

  init_ids ();
  parseconfig (configfile);
  mnt_init ();

  drop_privs ();
  if (!builddir)
    random_set_seedfile (strbuf ("%s/random_seed", sfsdir.cstr ()));
  random_init ();

  usrinfo_init ();
  afs_init (wrap (nullfn));
  unlink (sfsdevdb);

  amain ();
  return 0;
}
