/* $Id: usrinfo.C,v 1.39 1999/03/25 06:05:42 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */


#include "sfscd.h"
#include "sfsagent.h"

const time_t agent_timeout = 60;

ihash<const sfs_aid, usrinfo, &usrinfo::aid, &usrinfo::hlink> usrtab;

bool
isnfsmounter (const authunix_parms *aup)
{
  const sfs_aid aid = aup2aid (aup);
  return aid == sfsaid_sfs || aid == 0;
}

static void
nullccb_fn (clnt_stat)
{
}
static aclnt_cb nullccb (wrap (nullccb_fn));
static void
flushuser (sfs_aid aid, cdaemon *cdp)
{
  if (aclnt *c = cdp->c)
    c->call (SFSCDPROC_FLUSHAUTH, &aid, NULL, nullccb);
}

struct newagent {
  const sfs_aid aid;
  ref<axprt_stream> ax;
  ref<asrv> as;

  newagent (ref<axprt_stream> x, const authunix_parms *aup)
    : aid (aup2aid (aup)), ax (x),
      as (asrv::alloc (ax, agent_prog_1, wrap (this, &newagent::dispatch)))
    {}
  void dispatch (svccb *);
};

void
newagent::dispatch (svccb *sbp)
{
  if (!sbp) {
    delete this;
    return;
  }

  switch (sbp->proc ()) {
  case AGENT_NULL:
    sbp->reply (NULL);
    break;

  case AGENT_KILLSTART:
    {
      if (aid == sfsaid_sfs || aid == sfsaid_nobody || !aid)
	sbp->replyref (EINVAL);
      usrinfo *u;
      if ((u = usrtab[aid]))
	delete u;

      u = New usrinfo (aid);
      u->sh = as;
      u->ch = aclnt::alloc (ax, agentcb_prog_1);
      as->setcb (wrap (u, &usrinfo::dispatch));
      delete this;
      sbp->replyref (0);
      break;
    }

  case AGENT_START:
    if (aid == sfsaid_sfs || aid == sfsaid_nobody || !aid)
      sbp->replyref (EINVAL);
    else if (usrtab[aid])
      sbp->replyref (EBUSY);
    else {
      usrinfo *u = New usrinfo (aid);
      u->sh = as;
      u->ch = aclnt::alloc (ax, agentcb_prog_1);
      as->setcb (wrap (u, &usrinfo::dispatch));
      delete this;
      sbp->replyref (0);
    }
    break;

  case AGENT_KILL:
    if (usrinfo *u = usrtab[aid]) {
      delete u;
      sbp->replyref (0);
    }
    else
      sbp->replyref (ESRCH);
    break;

  default:
    sbp->replyref (EINVAL);
    break;
  }
}

usrinfo *
usrinfo::lookup (const authunix_parms *aup)
{
  return usrtab[aup2aid (aup)];
}

static void
authreq_cb (svccb *sbp, ref<sfsagent_auth_res> resp, clnt_stat err)
{
  if (err)
    resp->set_authenticate (false);
  sbp->reply (resp);
}

void
usrinfo::authreq (sfscd_authreq *ar, svccb *sbp)
{
  ref<sfsagent_auth_res> resp = New refcounted<sfsagent_auth_res>;
  switch (ar->type) {
  case AGENTCB_AUTHINIT:
    ch->timedcall (agent_timeout, AGENTCB_AUTHINIT, ar->init.addr (), resp,
		   wrap (authreq_cb, sbp, resp));
    break;
  case AGENTCB_AUTHMORE:
    ch->timedcall (agent_timeout, AGENTCB_AUTHMORE, ar->more.addr (), resp,
		   wrap (authreq_cb, sbp, resp));
    break;
  default:
    warn ("usrinfo::authreq: bad type\n");
    sbp->replyref (sfsagent_auth_res (false));
    break;
  }
}

afsdir *
userdir (const svccb *sbp)
{
  if (!sbp)
    return afs_naroot;
  const sfs_aid aid = aup2aid (sbp->getaup ());
  if (aid == sfsaid_sfs || !aid)
    return afs_sfsroot;
  if (usrinfo *u = usrtab[aid])
    return u->root;
  return afs_naroot;
}

static void
login_user (ref<afsdir> d, str name, ref<delaypt> dpt, struct srvinfo *si)
{
  // XXX - here we should log the user in
  d->unlink (name);
  if (afsnode *e = afs_sfsroot->lookup (name, NULL)) {
    d->link (e, name);
    dpt->setres (strbuf () << sfsroot << "/" << name);
  }
  else
    dpt->setres (NFSERR_NOENT);
}

static void
afsdir_unlink (ref<afsdir> d, str name)
{
  d->unlink (name);
}
static void
automount_cb (ref<afsdir> d, str name, ref<delaypt> dpt,
	      struct srvinfo *si, int err)
{
  if (si && !err)
    login_user (d, name, dpt, si);
  else {
    dpt->setres (NFSERR_NOENT);
    timecb (time (NULL) + 2, wrap (afsdir_unlink, d, name));
  }
}

void
agentlookup_cb (ref<afsusrdir> d, str name, ref<delaypt> dpt,
		sfsagent_lookup_res *resp, clnt_stat err)
{
  d->unlink (name);
  if (err || !resp->makelink || !resp->path->len ()) {
    d->mkulink (NULL, name);
    dpt->setres (NFSERR_NOENT);
  }
  else {
    d->mkulink (*resp->path, name);
    dpt->setres (*resp->path);
  }
  delete resp;
}

afsnode *
automount (svccb *sbp, str name)
{
  assert (!isnfsmounter (sbp->getaup ()));

  bool remote = sfs_parsepath (name);
  usrinfo *u = NULL;
  if (!remote) {
    u = usrinfo::lookup (sbp->getaup ());
    if (!nameok (name) || !u || u->root->negcache[name])
      return NULL;
  }

  srvinfo *si = remote ? srvinfo::lookup (name) : NULL;
  if (si && si->error)
    return NULL;

  ptr<afsdir> d = u ? implicit_cast<ref<afsdir> > (u->root)
    : mkref (userdir (sbp));
  d->unlink (name);
  ref<delaypt> dpt = delaypt::alloc ();
  d->link (dpt, name);

  if (remote) {
    if (si)
      login_user (d, name, dpt, si);
    else
      srvinfo::alloc (name, wrap (automount_cb, d, name, dpt));
  }
  else {
    sfsagent_lookup_res *resp = New sfsagent_lookup_res;
    u->ch->timedcall (agent_timeout, AGENTCB_LOOKUP, &name, resp,
		      wrap (agentlookup_cb, u->root, name, dpt, resp));
  }
  return dpt;
}

usrinfo::usrinfo (sfs_aid a)
  : aid (a), root (afsusrdir::alloc (afs_root))
{
  root->mkdir (strbuf (".id=%" U64F "u", a));
  usrtab.insert (this);
  cdtab.traverse (wrap (flushuser, aid));
}

usrinfo::~usrinfo ()
{
  usrtab.remove (this);
  cdtab.traverse (wrap (flushuser, aid));
}

void
usrinfo::dispatch (svccb *sbp)
{
  if (!sbp) {
    delete this;
    return;
  }

  switch (sbp->proc ()) {
  case AGENT_NULL:
    sbp->reply (NULL);
    break;
  case AGENT_KILLSTART:
    root->negcache.clear ();
    root = afsusrdir::alloc (afs_root);
    /* cascade */
  case AGENT_START:
    cdtab.traverse (wrap (flushuser, aid));
    sbp->replyref (0);
    break;
  case AGENT_KILL:
    sbp->replyref (0);
    delete this;
    break;
  case AGENT_SYMLINK:
    {
      sfsagent_symlink_arg *ssa
	= sbp->template getarg<sfsagent_symlink_arg> ();
      root->mkulink (ssa->contents, ssa->name);
      sbp->reply (NULL);
      break;
    }
  case AGENT_FLUSHNAME:
    {
      str name = *sbp->template getarg<sfs_hostname> ();
      root->clrulink (name);
      sbp->reply (NULL);
      break;
    }
  default:
    sbp->replyref (EINVAL);
    break;
  }
}

static void
usrinfo_accept (ptr<axprt_stream> ax, const authunix_parms *aup)
{
  if (!ax)
    fatal << "agent.sock" << ": " << strerror (errno) << "\n";
  assert (!ax->ateof ());
  vNew newagent (ax, aup);
}

void
usrinfo_init ()
{
  sfs_suidserv ("agent", wrap (usrinfo_accept));
}

static void
unlinkdir (str dirname, usrinfo *u)
{
  u->root->unlink (dirname);
}

void
flushpath (str path)
{
  afs_naroot->unlink (path);
  usrtab.traverse (wrap (unlinkdir, path));
}
