/* $Id: sfsconst.C,v 1.11 1999/05/27 23:44:44 kaminsky Exp $ */

/*
 *
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfsmisc.h"
#include "amisc.h"
#include "parseopt.h"
#include <pwd.h>
#include <grp.h>

#ifndef SFSUSER
# define SFSUSER "sfs"
#endif /* !SFSUSER */

u_int16_t sfs_port = SFS_PORT;
uid_t sfs_uid;
gid_t sfs_gid;
#ifdef MAINTAINER
bool runinplace;
#endif /* MAINTAINER */
str sfsdir = SFSDIR;
str sfssockdir = SFSDIR "/sockets";
str sfsdevdb = SFSDIR "/devdb";
const char *etc1dir = ETCDIR;
const char *etc2dir = SYSCONFDIR;
const char *etc3dir = NULL;
const char *sfsroot = "/sfs";

static bool const_set;

#ifdef MAINTAINER
# define fwarn warn
#else /* !MAINTAINER */
# define fwarn fatal
#endif /* !MAINTAINER */

static void
idlookup (str uid, str gid)
{
  if (!uid)
    uid = "sfs";
  if (!gid)
    gid = uid;

  bool uidok = convertint (uid, &sfs_uid);
  struct passwd *pw = uidok ? getpwuid (sfs_uid) : getpwnam (uid.cstr ());
  bool gidok = convertint (gid, &sfs_gid);
  struct group *gr = gidok ? getgrgid (sfs_gid) : getgrnam (gid.cstr ());

  if (!uidok) {
    if (!pw)
      fatal << "Could not find user " << uid << "\n";
    sfs_uid = pw->pw_uid;
  }
  if (!gidok) {
    if (!gr)
      fatal << "Could not find group " << gid << "\n";
    sfs_gid = gr->gr_gid;
  }
  if (gr && gr->gr_mem[0])
    fwarn << "Group " << gid << " must not have any members\n";
  if (pw && gr && (gid_t) pw->pw_gid != (gid_t) gr->gr_gid)
    fwarn << "User " << uid << " must have login group " << gid << ".\n";

  endpwent ();
  endgrent ();
}

static void
parseconfig (str cf)
{
  str uid, gid;
  bool dirset = false;

  bool errors = false;
  parseargs pa (cf);
  int line;
  vector<str> av;
  while (pa.getline (&av, &line)) {
    if (!strcasecmp (av[0], "sfsuser")) {
      if (uid) {
	errors = true;
	warn << cf << ":" << line << ": Duplicate sfsuser directive\n";
      }
      else if (av.size () == 2)
	uid = gid = av[1];
      else if (av.size () == 3) {
	uid = av[1];
	gid = av[2];
      }
      else
	warn << cf << ":" << line << ": usage: sfsuser user [group]\n";
    }
    else if (!strcasecmp (av[0], "sfsdir")) {
      if (dirset)
	warn << cf << ":" << line << ": Duplicate sfsdir directive\n";
      else if (av.size () != 2)
	warn << cf << ":" << line << ": usage: sfsuser user [group]\n";
      else if (!runinplace) {
	sfsdir = av[1];
	sfssockdir = sfsdir << "/sockets";
	sfsdevdb = sfsdir << "/devdb";
      }
      dirset = true;
    }
    else
      warn << cf << ":" << line << ": Unknown directive '"
	   << av[0] << "'\n";
  }
  if (errors)
    fatal << "errors in " << cf << "\n";

  idlookup (uid, gid);
}


void
sfsconst_init ()
{
  if (const_set)
    return;
  const_set = true;

#ifdef MAINTAINER
  if (char *p = getenv ("SFS_RUNINPLACE")) {
    runinplace = true;
    builddir = p;
  }
  if (char *p = getenv ("SFS_ROOT"))
    if (*p == '/')
      sfsroot = p;
#endif /* MAINTAINER */

#ifdef MAINTAINER
  if (runinplace) {
    sfsdir = builddir;
    sfssockdir = builddir;
    sfsdevdb = builddir << "/devdb";
    etc3dir = etc1dir;
    etc1dir = builddir;
    etc2dir = xstrdup (str (builddir << "/etc"));
  }
  if (char *ps = getenv ("SFS_PORT"))
    if (int pv = atoi (ps))
      sfs_port = pv;
#endif /* MAINTAINER */

  if (str cf = sfsconst_etcfile ("sfs_config"))
    parseconfig (cf);
  else
    idlookup ("sfs", NULL);

  if (!getuid () && !runinplace) {
    struct stat sb;
    mode_t m = umask (0);

    if (stat (sfsdir, &sb) < 0) {
      if (errno != ENOENT || (mkdir (sfsdir, 0755) < 0 && errno != EEXIST))
	fatal ("%s: %m\n", sfsdir.cstr ());
      if (chown (sfsdir, sfs_uid, sfs_gid) < 0) {
	int saved_errno = errno;
	rmdir (sfsdir);
	fatal ("chown (%s): %s\n", sfsdir.cstr (), strerror (saved_errno));
      }
      if (stat (sfsdir, &sb) < 0)
	fatal ("stat (%s): %m\n", sfsdir.cstr ());
    }
    if (sb.st_uid && sb.st_uid != sfs_uid)
      warn ("%s: owned by uid %d, should be uid %d\n",
	    sfsdir.cstr (), sb.st_uid, sfs_uid);
    if (sb.st_mode & 022)
      fwarn ("SECURITY HOLE -- %s has mode 0%o which should be 0755\n",
	     sfsdir.cstr (), sb.st_mode & 0777);

    if (stat (sfssockdir, &sb) < 0) {
      if (errno != ENOENT || (mkdir (sfssockdir, 0750) < 0 && errno != EEXIST))
	fatal ("%s: %m\n", sfssockdir.cstr ());
      if (chown (sfssockdir, sfs_uid, sfs_gid) < 0) {
	int saved_errno = errno;
	rmdir (sfssockdir);
	fatal ("chown (%s): %s\n", sfssockdir.cstr (), strerror (saved_errno));
      }
      if (stat (sfssockdir, &sb) < 0)
	fatal ("stat (%s): %m\n", sfssockdir.cstr ());
    }
    if (sb.st_uid && sb.st_uid != sfs_uid)
      warn ("%s: owned by uid %d, should be uid %d\n",
	    sfssockdir.cstr (), sb.st_uid, sfs_uid);
    if (sb.st_gid != sfs_gid)
      warn ("%s: group gid %d, should be gid %d\n",
	    sfssockdir.cstr (), sb.st_gid, sfs_gid);
    if (sb.st_mode & 027)
      fwarn ("SECURITY HOLE -- %s has mode 0%o which should be 0750\n",
	     sfssockdir.cstr (), sb.st_mode & 0777);

    umask (m);
  }
}

str
sfsconst_etcfile (const char *name)
{
  str file;
  file = strbuf ("%s/%s", etc1dir, name);
  if (!access (file, F_OK))
    return file;
  if (errno != ENOENT)
    fatal << file << ": " << strerror (errno) << "\n";
  file = strbuf ("%s/%s", etc2dir, name);
  if (!access (file, F_OK))
    return file;
  if (errno != ENOENT)
    fatal << file << ": " << strerror (errno) << "\n";
  if (etc3dir) {
    file = strbuf ("%s/%s", etc3dir, name);
    if (!access (file, F_OK))
      return file;
    if (errno != ENOENT)
      fatal << file << ": " << strerror (errno) << "\n";
  }
  return NULL;
}

str
sfsconst_etcfile_required (const char *name)
{
  str file = sfsconst_etcfile (name);
  str e3 = etc3dir ? str (strbuf ("  %s/%s\n", etc3dir, name)) : str ("");
  if (!file)
    fatal ("Could not find '%s'.  Searched:\n  %s/%s\n  %s/%s\n",
	   name, etc1dir, name, etc2dir, name) << e3;
  return file;
}
