/* $Id: sfs_prot.x,v 1.43 1999/05/14 08:07:37 dm Exp $ */

/*
 * This file was written by David Mazieres.  Its contents is
 * uncopyrighted and in the public domain.  Of course, standards of
 * academic honesty nonetheless prevent anyone in research from
 * falsely claiming credit for this work.
 */

%#include "bigint.h"

#ifdef SFSSVC
%#include "nfs3exp_prot.h"
#else /* !SFSSVC */
%#include "nfs3_prot.h"
const ex_NFS_PROGRAM = 344444;
const ex_NFS_V3 = 3;
#endif

const SFS_PORT = 4;
const SFS_RELEASE = 1;		/* 100 * release no. where protocol changed */

enum sfsstat {
  SFS_OK = 0,
  SFS_BADLOGIN = 1,
  SFS_NOSUCHHOST = 2,
  SFS_NOTSUPP = 10004,
  SFS_TEMPERR = 10008 
};

/* Types of hashed or signed messages */
enum sfs_msgtype {
  SFS_HOSTINFO = 1,
  SFS_KSC = 2,
  SFS_KCS = 3,
  SFS_SESSINFO = 4,
  SFS_AUTHINFO = 5,
  SFS_SIGNED_AUTHREQ = 6,
  SFS_AUTHREGISTER = 7,
  SFS_AUTHUPDATE = 8
};

/* Type of service requested by clients */
enum sfs_service {
  SFS_SFS = 1,			/* File system service */
  SFS_AUTHSERV = 2		/* Crypto key server */
};

typedef string sfs_extension<>;
typedef string sfs_hostname<255>;
typedef opaque sfs_hash[20];
typedef opaque sfs_secret[16];
typedef unsigned hyper sfs_seqno;

typedef bigint sfs_pubkey;
typedef bigint sfs_ctext;
typedef bigint sfs_sig;

struct sfs_hashcharge {
  unsigned int bitcost;
  sfs_hash target;
};
typedef opaque sfs_hashpay[64];


/*
 * Hashed structures
 */

/* Two, identical copies of of the sfs_hostinfo structure are
 * concatenated and then hashed with SHA-1 to form the hostid. */
struct sfs_hostinfo {
  sfs_msgtype type;		/* = SFS_HOSTINFO */
  sfs_hostname hostname;
  sfs_pubkey pubkey;
};

/* The two shared session keys, ksc and kcs, are the SHA-1 hashes of
 * sfs_sesskeydat with type = SFS_KCS or SFS_KSC.  */
struct sfs_sesskeydat {
  sfs_msgtype type;		/* = SFS_KSC or SFS_KCS */
  sfs_pubkey ks;
  sfs_secret sshare;		/* Server's share of session key */
  sfs_pubkey kc;
  sfs_secret cshare;		/* Client's share of session key */
};

/* The sessinfo structure is hashed to produce a session ID--a
 * structure both the client and server know to be fresh, but which,
 * unlike the session keys, can safely be divulged to 3rd parties
 * during user authentication.  */
struct sfs_sessinfo {
  sfs_msgtype type;		/* = SFS_SESSINFO */
  opaque ksc<>;			/* = SHA-1 ({SFS_KSC, ks, k1, kc, k3}) */
  opaque kcs<>;			/* = SHA-1 ({SFS_KCS, ks, k2, kc, k4}) */
};

/* The authinfo structure is hashed to produce an authentication ID.
 * The authentication ID can be computed by an untrusted party (such
 * as a user's unprivileged authentication agent), but allows that
 * third party to verify or log the hostname and hostid to which
 * authentication is taking place.  */
struct sfs_authinfo {
  sfs_msgtype type;		/* = SFS_AUTHINFO */
  sfs_service service;
  sfs_hostname name;
  sfs_hash hostid;		/* = SHA-1 (sfs_hostinfo, sfs_hostinfo) */
  sfs_hash sessid;		/* = SHA-1 (sfs_sessinfo) */
};

/*
 * Public key ciphertexts
 */

struct sfs_kmsg {
  sfs_secret kcs_share;
  sfs_secret ksc_share;
};

/*
 * Signed messages
 */

struct sfs_signed_authreq {
  sfs_msgtype type;		/* = SFS_SIGNED_AUTHREQ */
  sfs_hash authid;		/* SHA-1 (sfs_authinfo) */
  sfs_seqno seqno;		/* Counter, value unique per authid */
  opaque usrinfo[16];		/* All 0s, or <= 15 character logname */
};

/*
 * RPC arguments and results
 */

struct sfs_connectarg {
  unsigned int release;		/* = SFS_RELEASE */
  sfs_service service;
  sfs_hostname name;
  sfs_hash hostid;		/* = SHA1 (sfs_hostinfo, sfs_hostinfo) */
  sfs_extension extensions<>;
};

struct sfs_connectok {
  int release;			/* = SFS_RELEASE */
  sfs_hostinfo host;
  int prog;			/* Protocol used by this server */
  int vers;			/*    "      "   "   "     "    */
  sfs_hashcharge charge;
};
union sfs_connectres switch (sfsstat status) {
 case SFS_OK:
   sfs_connectok reply;
 default:
   void;
};

struct sfs_encryptarg {
  sfs_hashpay payment;
  sfs_ctext kmsg;
  sfs_pubkey pubkey;
};
typedef sfs_ctext sfs_encryptres;

struct sfs_nfs3_subfs {
  nfspath3 path;
  nfs_fh3 fh;
};
struct sfs_nfs3_fsinfo {
  nfs_fh3 root;
  sfs_nfs3_subfs subfs<>;
};
union sfs_nfs_fsinfo switch (int vers) {
 case ex_NFS_V3:
   sfs_nfs3_fsinfo v3;
};

union sfs_fsinfo switch (int prog) {
 case ex_NFS_PROGRAM:
   sfs_nfs_fsinfo nfs;
 default:
   void;
};


typedef string sfs_idname<32>;

union sfs_opt_idname switch (bool present) {
 case TRUE:
   sfs_idname name;
 case FALSE:
   void;
};

struct sfs_idnums {
  int uid;
  int gid;
};

struct sfs_idnames {
  sfs_opt_idname uidname;
  sfs_opt_idname gidname;
};

/*
 * User-authentication structures
 */

enum sfs_authtype {
  SFS_NOAUTH = 0,
  SFS_AUTHREQ = 1
};

struct sfs_authreq {
  sfs_pubkey usrkey;		/* Key with which signed_req signed */
  sfs_sig signed_req;		/* Recoveraby signed sfs_signed_authreq */
  /* string usrinfo<15>;	/* Logname or "" if any */
};

union sfs_autharg switch (sfs_authtype type) {
 case SFS_NOAUTH:
   void;
 case SFS_AUTHREQ:
   sfs_authreq req;
};

enum sfs_credtype {
  SFS_NOCRED = 0,
  SFS_UNIXCRED = 1
};

struct sfs_unixcred {
  string username<>;
  string homedir<>;
  string shell<>;
  unsigned uid;
  unsigned gid;
  unsigned groups<>;
};

union sfs_cred switch (sfs_credtype type) {
 case SFS_NOCRED:
   void;
 case SFS_UNIXCRED:
   sfs_unixcred unixcred;
};

struct sfs_loginokres {
  sfs_cred cred;
  sfs_hash authid;
  sfs_seqno seqno;
};

enum sfs_loginstat {
  SFSLOGIN_OK = 0,		/* Login succeeded */
  SFSLOGIN_MORE = 1,		/* More communication with client needed */
  SFSLOGIN_BAD = 2,		/* Invalid login */
  SFSLOGIN_ALLBAD = 3		/* Invalid login don't try again */
};

union sfsauth_loginres switch (sfs_loginstat status) {
 case SFSLOGIN_OK:
   sfs_loginokres resok;
 case SFSLOGIN_MORE:
   opaque resmore<>;
 case SFSLOGIN_BAD:
 case SFSLOGIN_ALLBAD:
   void;
};

union sfs_loginres switch (sfs_loginstat status) {
 case SFSLOGIN_OK:
   unsigned authno;
 case SFSLOGIN_MORE:
   opaque resmore<>;
 case SFSLOGIN_BAD:
 case SFSLOGIN_ALLBAD:
   void;
};

struct sfs_loginarg {
  sfs_seqno seqno;
  opaque certificate<>;		/* marshalled sfs_autharg */
};

/*
 * Secure Remote Password (SRP) protocol
 */

enum sfsauth_srpstat {
  SFSAUTH_SRP_OK = 0,
  SFSAUTH_SRP_NOTTHERE = 1,
  SFSAUTH_SRP_PHASE0 = 2,
  SFSAUTH_SRP_PHASE1 = 3,
  SFSAUTH_SRP_PHASE2 = 4,
  SFSAUTH_SRP_BADSRPINFO = 5
};

struct sfsauth_srpinfo {
  string prime<>;
  string generator<>;
  string salt<>;
  string verifier<>;
  string seckey<>;
};

/* The key field is determined by the SRP protocol and is known by 
 * both the client and the server.  The session key for each direction
 * is determined by hashing sfsauth_sesskeydat with type = SFS_KCS or 
 * SFS_KSC.  */
struct sfsauth_sesskeydat {
  sfs_msgtype type;		/* = SFS_KSC or SFS_KCS */
  opaque key<>;			/* from SRP protocol */
};

struct sfsauth_srp0arg {
  string username<>;
};

struct sfsauth_srp0resok {
  string prime<>;
  string generator<>;
};

union sfsauth_srp0res switch (sfsauth_srpstat status) {
 case SFSAUTH_SRP_OK:
   sfsauth_srp0resok resok;
 default:
   void;
};

struct sfsauth_srp1arg {
  string A<>;
};

struct sfsauth_srp1resok {
  string salt<>;
  string B<>;
  string u<>;
};

union sfsauth_srp1res switch (sfsauth_srpstat status) {
 case SFSAUTH_SRP_OK:
   sfsauth_srp1resok resok;
 default:
   void;
};

struct sfsauth_srp2arg {
  sfs_hash clntproof;
};

struct sfsauth_srp2resok {
  sfs_hash servproof;
};

union sfsauth_srp2res switch (sfsauth_srpstat status) {
 case SFSAUTH_SRP_OK:
   sfsauth_srp2resok resok;
 default:
   void;
};

enum sfsauth_fetchstat {
  SFSAUTH_FETCH_OK = 0,
  SFSAUTH_FETCH_FAILED = 1
};

struct sfsauth_fetchresok {
  string seckey<>;
  sfs_hash hostid;
};

union sfsauth_fetchres switch (sfsauth_fetchstat status) {
 case SFSAUTH_FETCH_OK:
   sfsauth_fetchresok resok;
 default:
   void;
};

struct sfsauth_registerarg {
  string username<>;		/* logname */
  string password<>;		/* password for an add */
  sfs_pubkey pubkey;
  sfsauth_srpinfo *srpinfo;
};

enum sfsauth_registerres {
  SFSAUTH_REGISTER_OK = 0,
  SFSAUTH_REGISTER_NOTSOCK = 1,
  SFSAUTH_REGISTER_BADUSERNAME = 2,
  SFSAUTH_REGISTER_WRONGUID = 3,
  SFSAUTH_REGISTER_DENYROOT = 4,
  SFSAUTH_REGISTER_BADSHELL = 5,
  SFSAUTH_REGISTER_DENYFILE = 6,
  SFSAUTH_REGISTER_BADPASSWORD = 7,
  SFSAUTH_REGISTER_USEREXISTS = 8,
  SFSAUTH_REGISTER_FAILED = 9
};

struct sfsauth_updatereq {
  sfs_msgtype type;		/* = SFS_AUTHUPDATE */
  sfs_hash authid;		/* SHA-1 (sfs_authinfo);
				   service is SFS_AUTHSERV */
  sfs_pubkey oldkey;
  sfs_pubkey newkey;
  sfsauth_srpinfo *srpinfo;
  /* maybe username? */
};

struct sfsauth_updatearg {
  sfsauth_updatereq updatereq;
  sfs_sig signature;		/* computed with sfsauth_updatereq.oldkey */
};

enum sfsauth_updateres {
  SFSAUTH_UPDATE_OK = 0,
  SFSAUTH_UPDATE_BADSIGNATURE = 1,
  SFSAUTH_UPDATE_PROTOERR = 2,
  SFSAUTH_UPDATE_NOTTHERE = 3,
  SFSAUTH_UPDATE_BADAUTHID = 4,
  SFSAUTH_UPDATE_DENYFILE = 5,
  SFSAUTH_UPDATE_FAILED = 6
};

program SFS_PROGRAM {
	version SFS_VERSION {
		void 
		SFSPROC_NULL (void) = 0;

		sfs_connectres
		SFSPROC_CONNECT (sfs_connectarg) = 1;

		sfs_encryptres
		SFSPROC_ENCRYPT (sfs_encryptarg) = 2;

		sfs_fsinfo
		SFSPROC_GETFSINFO (void) = 3;

		sfs_loginres
		SFSPROC_LOGIN (sfs_loginarg) = 4;

		void
		SFSPROC_LOGOUT (unsigned) = 5;

		sfs_idnames
		SFSPROC_IDNAMES (sfs_idnums) = 6;

		sfs_idnums
		SFSPROC_IDNUMS (sfs_idnames) = 7;

		sfs_cred
		SFSPROC_GETCRED (void) = 8;
	} = 1;
} = 344440;

program SFSCB_PROGRAM {
	version SFSCB_VERSION {
		void 
		SFSCBPROC_NULL(void) = 0;
	} = 1;
} = 344441;

program SFSAUTH_PROGRAM {
	version SFSAUTH_VERSION {
		void 
		SFSAUTHPROC_NULL (void) = 0;

		sfsauth_loginres
		SFSAUTHPROC_LOGIN (sfs_loginarg) = 1;

		sfsauth_registerres
		SFSAUTHPROC_REGISTER (sfsauth_registerarg) = 2;

		sfsauth_updateres
		SFSAUTHPROC_UPDATE (sfsauth_updatearg) = 3;

		sfsauth_srpinfo
		SFSAUTHPROC_SRP_GETGLOBALPARAMS (void) = 4;

		sfsauth_srp0res
		SFSAUTHPROC_SRP_PHASE0 (sfsauth_srp0arg) = 5;

		sfsauth_srp1res
		SFSAUTHPROC_SRP_PHASE1 (sfsauth_srp1arg) = 6;

		sfsauth_srp2res
		SFSAUTHPROC_SRP_PHASE2 (sfsauth_srp2arg) = 7;

		sfsauth_fetchres
		SFSAUTHPROC_FETCH (void) = 8;
	} = 1;
} = 344442;
