/* $Id: inode.c,v 1.2 1999/03/24 11:07:25 kaminsky Exp $ */

/*
 * User level VFS driver for Linux.
 * Copyright 1999 Michael Kaminsky <kaminsky@lcs.mit.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#define UVFS_DEBUG_FS 1

#include "uvfs_kern.h"
#include <linux/locks.h>

static struct super_block *uvfs_read_super (struct super_block *sb,
					     void *data, int silent);

struct file_system_type uvfs_fs_type = {
  MOUNT_UVFS,
  0,
  uvfs_read_super,
  NULL
};

static void uvfs_put_super (struct super_block *sb);
static int uvfs_statfs (struct super_block *sb, struct statfs *buf, 
			int bufsiz);
static void uvfs_read_inode (struct inode *inode);

static struct super_operations uvfs_super_ops = {
  uvfs_read_inode,		/* read_inode */
  NULL,				/* write_inode */
  NULL,				/* put_inode */
  NULL,				/* delete_indo */
  NULL,				/* notify_change */
  uvfs_put_super,
  NULL,				/* write_super */
  uvfs_statfs,
  NULL,
  NULL,
  NULL
};

int
uvfs_newinode (struct super_block *sb, struct inode **ipp, unsigned long ino)
{
  struct inode *ip;

  ip = iget (sb, ino);
  if (ip == NULL)
    return -ENOMEM;
  ITOUVFS(ip) = xmalloc (sizeof (uvfs_node));
  if (ITOUVFS(ip) == NULL)
    return -ENOMEM;
  ITOUVFS(ip)->uvfslockf = NULL;
  ITOUVFS(ip)->uvfsinode = ip;
  *ipp = ip;
  return 0;
}

static struct super_block *
uvfs_read_super (struct super_block *sb, void *data, int silent)
{
  struct uvfs_args *args = (struct uvfs_args *) data;
  struct inode *root_inode;
  uvfs_mntpt *mntpt;
  int i, error;

  MOD_INC_USE_COUNT;

#if UVFS_DEBUG_FS
  warn ("uvfs_read_super: \n");
#endif /* UVFS_DEBUG_FS */

  if (!args) {
    warn ("uvfs_read_super: missing data arguments to mount\n");
    sb->s_dev = 0;
    MOD_DEC_USE_COUNT;
    return NULL;
  }

  warn ("uvfs_read_super: args->uvfs_dev = %u; args->uvfs_root_fh = %u\n",
	args->uvfs_dev, args->uvfs_root_fh);

  if ((size_t) args->uvfs_dev >= NUVFS) {
    warn ("uvfs_read_super: uvfs device minor number out of range\n");
    sb->s_dev = 0;
    MOD_DEC_USE_COUNT;
    return NULL;
  }

/*   if (mp->mnt_flag & MNT_UPDATE) */
/*     return EOPNOTSUPP; */
/*   mp->mnt_flag |= MNT_NOSUID | MNT_NODEV; */

  lock_super (sb);

  sb->s_magic = VT_UVFS;
  sb->s_op = &uvfs_super_ops;
  sb->s_blocksize = 1024;	/* XXX: Is this OK? */
  sb->s_blocksize_bits = 10;	/* XXX: Is this OK? */

  mntpt = xmalloc (sizeof (uvfs_mntpt));
  if (mntpt == NULL) {
    warn ("uvfs_read_super: couldn't allocate mntpt struct\n");
    sb->s_dev = 0;
    unlock_super (sb);
    MOD_DEC_USE_COUNT;
    return NULL;
  }

  SBTOUVFS(sb) = (void *) mntpt;
  mntpt->rpcqp = &uvfs_state[args->uvfs_dev].rpcq;
  for (i = 0; i < UNTSIZE; i++) {
    LIST_INIT(&mntpt->uvfs_node_tab[i]);
  }

  error = uvfs_newinode (sb, &root_inode, args->uvfs_root_fh);
  if (error) {
    warn ("uvfs_read_super: couldn't get root inode\n");
    xfree (mntpt);
    sb->s_dev = 0;
    unlock_super (sb);
    MOD_DEC_USE_COUNT;
    return NULL;
  }

  root_inode->i_mode |= S_IFDIR;
  root_inode->i_op = &uvfs_dir_inode_operations;
  mntpt->root = root_inode;

  sb->s_root = d_alloc_root (mntpt->root, NULL);
  if (!sb->s_root) {
    warn ("uvfs_read_super: couldn't allocate root dentry\n");
    iput (mntpt->root);
    xfree (mntpt);
    sb->s_dev = 0;
    unlock_super (sb);
    MOD_DEC_USE_COUNT;
    return NULL;
  }
  sb->s_root->d_op = &uvfs_dentry_operations;

  unlock_super (sb);
  return sb;
}

static void
uvfs_put_super (struct super_block *sb)
{
  uvfs_mntpt *mntpt = SBTOUVFS(sb);

#if UVFS_DEBUG_FS
  warn ("uvfs_umount: \n");
#endif /* UVFS_DEBUG_FS */

  lock_super (sb);
  sb->s_dev = 0;
  unlock_super (sb);
  xfree (mntpt);
  MOD_DEC_USE_COUNT;
  return;
}

static int uvfs_statfs (struct super_block *sb, struct statfs *buf, 
			int bufsiz)
{
  struct statfs tmp;
  int error = 0;

#if UVFS_DEBUG_FS
  warn ("uvfs_statfs: \n");
#endif /* UVFS_DEBUG_FS */

  tmp.f_type = sb->s_magic;
  tmp.f_bsize = sb->s_blocksize;
  tmp.f_blocks = 21144;
  tmp.f_bfree = 9872;
  tmp.f_bavail = tmp.f_bfree;
  tmp.f_files = 122;
  tmp.f_ffree = 12;
  tmp.f_namelen = NAME_MAX;
  error = copy_to_user (buf, &tmp, bufsiz);
  warn ("statfs: error = %d\n", error);
  if (error)
    return -EFAULT;
  else
    return 0;
}

static void 
uvfs_read_inode (struct inode *ip)
{
  return;
}
